#!/usr/bin/env python3
"""
Author: Chris Dijkstra
Date: 11/10/2023

Utility functions for image analysis
"""
import numpy as np
import skimage as sk


def crop_region(
        image: np.ndarray, centre: int, shape: tuple[int, int]
) -> np.ndarray:
    """ Crops an image area of specified width and height around a central point

    :param image: matrix representing the image
    :param centre: x and y coordinate of the centre as integers
    :param shape: contains the height and width of the subregion in pixels as
        integers
    :return: The cropped region of the original image
    """
    if image.ndim == 2:
        crop = image[
               centre[1] - shape[0] // 2: centre[1] + shape[0] // 2,
               centre[0] - shape[1] // 2: centre[0] + shape[1] // 2
               ]
    else:
        crop = image[
               centre[1] - shape[0] // 2: centre[1] + shape[0] // 2,
               centre[0] - shape[1] // 2: centre[0] + shape[1] // 2,
               :
               ]
    return crop


def read_fimg(filename: str) -> np.ndarray:
    """ Turns an FIMG value into a normalized file with data between 0 and 1

    :param filename: name of the file that is to be opened
    :return: 2D array representing the fimg image
    """
    image = np.fromfile(filename, np.dtype("float32"))
    image = image[2:]
    image = np.reshape(image, newshape=(1024, 1360))
    image[image < 0] = 0
    return image


def increase_contrast(im_channel: np.ndarray) -> np.ndarray:
    """ Takes a 2d array and makes its values range from 0 to 1

    :param im_channel: numpy array with 2 dimensions
    :return: input channel scaled from 0 to 1.
    """
    ch_min = im_channel.min()
    ch_max = im_channel.max()
    if ch_max == ch_min:
        raise ValueError("All values in the image are identical.")
    out = (im_channel - ch_min) / (ch_max - ch_min)
    return out


def multichannel_mask(image: np.ndarray, mask: np.ndarray) -> np.ndarray:
    """ Takes an image and applies a mask to every channel

    :param image: 3 dimensional array representing an image
    :param mask: 2d binary mask
    :return: masked input image
    """
    mask = mask.astype(image.dtype)
    image = image.copy()
    image[:, :, 0] *= mask
    image[:, :, 1] *= mask
    image[:, :, 2] *= mask
    return image


def paint_col(
        image: np.ndarray, mask: np.ndarray, color_tuple: tuple[int, int, int]
) -> np.ndarray:
    """ Makes masked area the specified color

    :param image: 3d array representing an RGB image
    :param mask: 2d binary mask
    :param color_tuple: contains the values in integer of the R, G and B
        channel that you want to paint
    :return: same as input image but with the masked area painted in
        the specified color
    """
    image = image.copy()
    image[:, :, 0][mask == 0] = color_tuple[0]
    image[:, :, 1][mask == 0] = color_tuple[1]
    image[:, :, 2][mask == 0] = color_tuple[2]
    return image


def merge_masks(bg_mask: np.ndarray, pheno_mask: np.ndarray) -> np.ndarray:
    """ Merges 2 binary masks into one mask, where phenotype has high values

    :param bg_mask: 2D mask with background as 0 and foreground as
        1
    :param pheno_mask: 2D mask phenotype marked as 1 and everything
        else as 0
    :return: 2D mask with background marked as 0, foreground as 1 and
        phenotype area as 2
    """
    substep = bg_mask.astype(int) + pheno_mask.astype(int)
    comb_mask = np.zeros_like(bg_mask)
    comb_mask[substep == 1] = 2
    comb_mask[substep == 2] = 1
    comb_mask[bg_mask == 0] = 0
    return comb_mask


def threshold_between(
        image: np.ndarray, x_low: float | int = None,
        x_high: float | int = None, y_low: float | int = None,
        y_high: float | int = None, z_low: float | int = None,
        z_high: float | int = None, and_mask: bool = True
) -> np.ndarray:
    """ Thresholds an image array for being between two values for each channels

    :param image: 3d matrix representing the image
    :param x_low: low boundary for channel 1. Defaults to minimum of channel 1.
    :param x_high: high boundary for channel 1. Defaults to maximum of channel
        1.
    :param y_low: low boundary for channel 2. Defaults to minimum of channel 2.
    :param y_high: high boundary for channel 2. Defaults to maximum of channel
        2.
    :param z_low: low boundary for channel 3. Defaults to minimum of channel 3.
    :param z_high: high boundary for channel 3. Defaults to maximum of channel
        3.
    :param and_mask: if true returned mask is only true when all
        thresholds apply. If false returned mask is true if at least one of the
        thresholds apply.
    :return: binary mask
    """
    # channel x thresholding
    if not x_low:
        x_low = image[:, :, 0].min()
    if not x_high:
        x_high = image[:, :, 0].max()
    x_mask = (image[:, :, 0] >= x_low) & (image[:, :, 0] <= x_high)
    # channel y thresholding
    if not y_low:
        y_low = image[:, :, 1].min()
    if not y_high:
        y_high = image[:, :, 1].max()
    y_mask = (image[:, :, 1] >= y_low) & (image[:, :, 1] <= y_high)
    # channel z thresholding
    if not z_low:
        z_low = image[:, :, 2].min()
    if not z_high:
        z_high = image[:, :, 2].max()
    z_mask = (image[:, :, 2] >= z_low) & (image[:, :, 2] <= z_high)
    if and_mask:
        comp_mask = x_mask & y_mask & z_mask
    else:
        comp_mask = x_mask | y_mask | z_mask
    return comp_mask


def label_horizontally(mask: np.ndarray, left_right: bool = True) -> np.ndarray:
    """ Labels images from left to right

    :param mask: Binary mask of objects that need to be labelled
    :param left_right: If true labels left to right, if false right to left
    :return: Mask of labelled objects
    """
    rotated = np.rot90(mask, 3) if left_right else np.rot90(mask, 1)
    labeled = sk.measure.label(rotated)
    reset = np.rot90(labeled, 1) if left_right else np.rot90(labeled, 3)
    return reset


def parse_coord_file(filename: str) -> list[tuple[int, int]]:
    """ Parser for ImageJ xy.txt output files

    :param filename: The input file
    :return: List where each element is a tuple of ints in form (x, y)
    """""
    out_lines = []
    with open(filename) as infile:
        for line in infile:
            line = line.strip().split(sep="\t")
            line = [int(float(elem)) for elem in line]
            out_lines.append(tuple(line))
    return out_lines



"""
Author: Chris Dijkstra
Date: 11-11-2024

Functions for creating distance maps of masked images where each pixel gets assigned a value based on their proximity
to the nearest seed.
"""

import numpy as np
import skimage as sk
import scipy.ndimage as ndi


def coord_dist_map(
        mask: np.ndarray, seed_coords: list[tuple[int, int]]
) -> np.ndarray:
    """ Creates a distance map from seeds given as coordinates, bound by a mask

    :param mask: The binary mask that gives the boundaries of the distance map
    :param seed_coords: The list of coordinates
    :return: A distance map
    """
    seed_map = np.zeros_like(mask)
    for x, y in seed_coords:
        seed_map[y, x] = 1
    dist_map = im_dist_map(mask == 1, seed_map)
    return dist_map


def im_dist_map(mask: np.ndarray, seed_map: np.ndarray[bool]) -> np.ndarray:
    """ Creates a distance map from seeds derived from an input image, bound by a mask

    :param mask: The binary mask that gives the boundaries of the distance map
    :param seed_map: Binary image containing the seeds as 1 or True
    :return: A distance map
    """
    seed_map = np.invert(seed_map)
    seed_map[mask == 0] = 1

    if (seed_map == 0).sum() == 0:
        dist_map = np.zeros_like(mask)
    else:
        dist_map = ndi.distance_transform_edt(seed_map)

    dist_map[mask == 0] = 0
    return dist_map


def centre_seed_dist(flip_mask: np.ndarray[bool], non_flip_mask: np.ndarray[bool], seed_map: np.ndarray[bool]) -> \
        np.ndarray:
    """ Creates a dist map from a line bisecting a mask and transforms it into a continuous distance map

    :param flip_mask: Mask of the area of the distance map that will be flipped. The pixel that is furthest from the
        bisecting line in seed_map will be 0 in the output map
    :param non_flip_mask: Mask of the area of the distance map that will not be flipped
    :param seed_map: The seed map containing the line that bisects the mask into flip_mask and non_flip_mask, should
        have overlap with both flip_mask and non_flip_mask
    :return: A distance map
    """
    dist_map_f = im_dist_map(flip_mask, seed_map)
    dist_map_f_inverted = abs(dist_map_f - np.max(dist_map_f))
    dist_map_f_inverted[flip_mask == 0] = 0

    dist_map_nf = im_dist_map(non_flip_mask, seed_map)
    dist_map_nf += np.max(dist_map_f_inverted)
    dist_map_nf[non_flip_mask == 0] = 0

    dist_map = dist_map_f_inverted + dist_map_nf
    return dist_map



from setuptools import setup, find_packages

# read the contents of README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="lettuceSee",
    url="https://github.com/Chizz98/LettuceSee",
    packages=find_packages(),
    author="Chris Dijkstra",
    author_email="chris_dijkstra98@hotmail.com",
    description="A package of image analysis algorithms suited for plants",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="0.1.0",
    install_requires=[
        "scikit-image",
        "scipy",
        "numpy"
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Intended Audience :: Science/Research"
    ]
)



#!/usr/bin/env python3
"""
Author: Chris Dijkstra
Date: 10/10/2023

Contains functions for segmenting image arrays.
"""
import skimage as sk
import numpy as np
from scipy import signal
from . import util


def elevation_map(rgb_im: np.ndarray) -> np.ndarray:
    """ Creates an elevation map of an RGB image based on sobel filtering

    :param rgb_im: 3 dimensional array representing an RGB image
    :return: 2 dimensional array representing an edge map
    """
    compound_sobel = sk.filters.sobel(rgb_im)
    compound_sobel = compound_sobel[:, :, 0] + compound_sobel[:, :, 1] + \
                     compound_sobel[:, :, 2]
    elevation = sk.filters.sobel(compound_sobel)
    return elevation


def multichannel_threshold(
        multi_ch_im, x_th: float = 0.0, y_th: float = 0.0, z_th: float = 0.0,
        inverse: bool = False
) -> np.ndarray:
    """ Takes a three-channel image and returns a mask based on thresholds

    :param multi_ch_im: a numpy array representing an image with
        three color channels
    :param x_th: the threshold for the first channel, 0.0 by default
    :param y_th: the threshold for the second channel, 0.0 by default
    :param z_th: the threshold for the third channel, 0.0 by default
    :param inverse: if False pixels below the threshold are marked as 0,
        if True, pixels above the threshold are marked as 0.
    :return: the mask created based on the thresholds, 2D array
        same width and height as the input
    """
    mask = np.ones(multi_ch_im.shape[0:2])
    mask[multi_ch_im[:, :, 0] < x_th] = 0
    mask[multi_ch_im[:, :, 1] < y_th] = 0
    mask[multi_ch_im[:, :, 2] < z_th] = 0
    mask = mask.astype(int)
    if inverse:
        mask = np.invert(mask)
    return mask


def sh_markers(image: np.ndarray, distance: int = 10, bg_mod: float = 0.15,
               fg_mod: float = 0.2) -> np.ndarray:
    """ Creates marker image based on sobel histogram

    :param image: 3d array representing a 3 channel image
    :param distance: minimal distance between local maxima and minima
    :param bg_mod: modifier for histogram segmentation
    :param fg_mod: modifier for histogram segmentation
    :return: 2D marker image
    """
    if image.shape[2] == 4:
        image = sk.util.img_as_ubyte(sk.color.rgba2rgb(image))
    comp_sob = sk.filters.sobel(image)
    comp_sob = comp_sob[:, :, 0] + comp_sob[:, :, 1] + comp_sob[:, :, 2]
    values, bins = np.histogram(comp_sob, bins=100)
    max_i, _ = signal.find_peaks(values, distance=distance)
    max_bins = bins[max_i]
    min_i, _ = signal.find_peaks(-values, distance=distance)
    min_bins = bins[min_i]
    min_bins = min_bins[min_bins > max_bins[0]]
    markers = np.zeros_like(comp_sob)
    markers[
        comp_sob <= max_bins[0] + (bg_mod * (min_bins[0] - max_bins[0]))] = 1
    markers[
        comp_sob >= min_bins[0] + (fg_mod * (max_bins[1] - min_bins[0]))] = 2
    markers = markers.astype(int)
    return markers


def shw_segmentation(
        image: np.ndarray, distance: int = 10, bg_mod: float = 0.15,
        fg_mod: float = 0.2
) -> np.ndarray:
    """ Creates binary image through sobel + histogram thresholds + watershed

    :param image: array representing a 3d image
    :param distance: minimal distance between local maxima and minima
    :param bg_mod: modifier for histogram segmentation
    :param fg_mod: modifier for histogram segmentation
    :return: 2D mask for the image
    """
    if image.shape[2] == 4:
        image = sk.util.img_as_ubyte(sk.color.rgba2rgb(image))
    comp_sob = sk.filters.sobel(image)
    comp_sob = comp_sob[:, :, 0] + comp_sob[:, :, 1] + comp_sob[:, :, 2]
    elevation = sk.filters.sobel(comp_sob)
    values, bins = np.histogram(comp_sob, bins=100)
    max_i, _ = signal.find_peaks(values, distance=distance)
    max_bins = bins[max_i]
    min_i, _ = signal.find_peaks(-values, distance=distance)
    min_bins = bins[min_i]
    min_bins = min_bins[min_bins > max_bins[0]]
    markers = np.zeros_like(comp_sob)
    markers[
        comp_sob <= max_bins[0] + (bg_mod * (min_bins[0] - max_bins[0]))] = 1
    markers[
        comp_sob >= min_bins[0] + (fg_mod * (max_bins[1] - min_bins[0]))] = 2
    markers = markers.astype(int)
    mask = sk.segmentation.watershed(elevation, markers)
    mask = sk.morphology.erosion(mask, footprint=sk.morphology.disk(2))
    return mask - 1


def barb_thresh(im_channel: np.ndarray, div: int = 3) -> float:
    """ Defines the threshold of an image channel based on its histogram

    :param im_channel: 2d array, meant to be hue channel of hsv or
        a channel of lab
    :param div: the divisor used at the end of the algorithm. A higher
        divisor will lead to a lower threshold
    :return: the threshold of the image channel that separates it into
        healthy and unhealthy tissue
    """
    values, bins = np.histogram(im_channel, bins=100)
    peak_i = np.argmax(values)
    val_max = values[peak_i]
    bin_max = bins[peak_i]
    if bin_max <= 0.4:
        bound = 0.2 * val_max
    else:
        bound = 0.5 * val_max
    ref_val = values[values > bound][-1]
    ref_i = np.where(values == ref_val)[0][0]
    ref_bin = bins[ref_i]
    thresh = 2 * ref_bin / div
    return thresh


def barb_hue(
        image: np.ndarray, bg_mask: np.ndarray = None, div: int = 3
) -> np.ndarray:
    """ Takes an image of plant tissue and segments into healthy and brown

    :param image: 3d array representing an rgb image
    :param bg_mask: 2d array to mask the background
    :param div: the divisor used at the end of the algorithm. A higher
        divisor will lead to a lower threshold
    :return: mask with background as 0, healthy tissue as 1 and
        brown tissue as 2
    """
    if bg_mask is not None:
        # Apply mask to rgb_im
        image = util.multichannel_mask(image, bg_mask)
    # Get hue channel and scale from 0 to 1
    hue = sk.color.rgb2hsv(image)[:, :, 0]
    hue_con = util.increase_contrast(hue)
    hue_fg = hue_con[bg_mask == 1]
    # Healthy tissue masking
    thresh = barb_thresh(hue_fg, div)
    healthy_mask = (hue_con > thresh).astype(int)
    # Remove noise
    healthy_mask = sk.morphology.remove_small_holes(
        healthy_mask,
        area_threshold=(image.shape[0] + image.shape[1]) // 200
    )
    # Combine healthy and bg mask to get compound image
    bg_mask = bg_mask.astype(int)
    comp_mask = util.merge_masks(bg_mask, healthy_mask)
    return comp_mask


def canny_labs(image: np.ndarray, mask: np.ndarray, sigma: float) -> np.ndarray:
    """ Separates objects trough canny lines and then labels the output

    :param image: 2d array representing an image
    :param mask: 2d binary mask
    :param sigma: the sigma used for the gaussian blur component of canny
        segmentation
    :return: labelled image
    """
    canny_f = sk.feature.canny(image, sigma=sigma)
    canny_f = sk.morphology.closing(
        canny_f, footprint=sk.morphology.disk(radius=min(image.shape) // 500)
    )
    canny_f = sk.morphology.skeletonize(canny_f)
    mask = mask.copy()
    mask[canny_f == 1] = 0
    labels = sk.measure.label(mask, connectivity=1)
    return sk.morphology.dilation(labels)


def centre_primary_label(
        lab_im: np.ndarray, edge_length: int = 200, bg_label: int = 0
) -> np.ndarray:
    """ Takes labelled image and returns the label of the central object

    :param lab_im: labelled image with only positive values and 0
    :param edge_length: height and width of the square used on the centre
    :param bg_label: the label number that will be considered background.
        This can not be chosen as the primary label.
    :return: primary label
    """
    centre = (lab_im.shape[0] // 2, lab_im.shape[1] // 2)
    crop = util.crop_region(lab_im, centre, (edge_length, edge_length))
    return np.argmax(np.bincount(crop[crop != bg_label].ravel()))


def central_ob(mask: np.ndarray, edge_length: int = 200) -> np.ndarray:
    """ Selects the largest object in the centre of the image

    :param mask: 2d array representing a background mask
    :param edge_length: height and width of the square considered the
        centre
    :return: 2d array representing the input mask with only the
        largest central object
    """
    label_image = sk.measure.label(mask)
    central_lab = centre_primary_label(label_image, edge_length)
    centre_mask = (label_image == central_lab).astype(int)
    return centre_mask


def canny_central_ob(
        image: np.ndarray, mask: np.ndarray, sigma: float,
        central_area: int = 200
) -> np.ndarray:
    """ Uses canny filter and color channel thresholding to take central object

    :param image: 3d array representing rgb image
    :param mask: 2d boolean array representing background mask
    :param sigma: sigma used for gaussian blur step of canny edge
        detection
    :param central_area: central area size
    :return: 2d binary mask of central object
    """
    bg_labs = sk.measure.label(mask)
    mask = bg_labs == centre_primary_label(bg_labs)
    canny_labelled = canny_labs(sk.color.rgb2gray(image), mask, sigma)
    prim_lab = centre_primary_label(canny_labelled, central_area)
    average_cols = sk.color.label2rgb(canny_labelled, image, kind="avg")
    average_cols = sk.color.rgb2hsv(average_cols)
    prim_area = util.multichannel_mask(average_cols, canny_labelled == prim_lab)
    h_main = np.unique(prim_area[:, :, 0])[1]
    s_main = np.unique(prim_area[:, :, 1])[1]
    v_main = np.unique(prim_area[:, :, 2])[1]
    mask = util.threshold_between(
        image=average_cols,
        x_low=h_main - 0.1, x_high=h_main + 0.1,
        y_low=s_main - 0.25, y_high=s_main + 0.25,
        z_low=v_main - 0.25, z_high=v_main + 0.25
    )
    mask = sk.morphology.closing(mask, footprint=sk.morphology.disk(3))
    mask = sk.morphology.remove_small_holes(mask, area_threshold=150)
    return mask
